import collections
from typing import Any, Sequence, TypeAlias

import keras
import tensorflow as tf

from keras_rs.src import types
from keras_rs.src.layers.embedding import distributed_embedding_config

FeatureConfig = distributed_embedding_config.FeatureConfig
TableConfig = distributed_embedding_config.TableConfig

# Placeholder of tf.tpu.experimental.embedding._Optimizer which is not exposed.
TfTpuOptimizer: TypeAlias = Any


OptimizerMapping = collections.namedtuple(
    "OptimizerMapping",
    ["tpu_optimizer_class", "supported_kwargs", "unsupported_kwargs"],
)


OPTIMIZER_MAPPINGS = {
    keras.optimizers.Adagrad: OptimizerMapping(
        tpu_optimizer_class=tf.tpu.experimental.embedding.Adagrad,
        supported_kwargs=["initial_accumulator_value"],
        unsupported_kwargs={"epsilon": 1e-07},
    ),
    keras.optimizers.Adam: OptimizerMapping(
        tpu_optimizer_class=tf.tpu.experimental.embedding.Adam,
        supported_kwargs=["beta_1", "beta_2", "epsilon"],
        unsupported_kwargs={"amsgrad": False},
    ),
    keras.optimizers.Ftrl: OptimizerMapping(
        tpu_optimizer_class=tf.tpu.experimental.embedding.FTRL,
        supported_kwargs=[
            "learning_rate_power",
            "initial_accumulator_value",
            "l1_regularization_strength",
            "l2_regularization_strength",
            "beta",
        ],
        unsupported_kwargs={"l2_shrinkage_regularization_strength": 0.0},
    ),
    keras.optimizers.SGD: OptimizerMapping(
        tpu_optimizer_class=tf.tpu.experimental.embedding.SGD,
        supported_kwargs=[],
        unsupported_kwargs={"momentum": 0.0, "nesterov": False},
    ),
}


# KerasRS to TensorFlow


def keras_to_tf_tpu_configuration(
    feature_configs: types.Nested[FeatureConfig],
    table_stacking: str | Sequence[str] | Sequence[Sequence[str]],
    num_replicas_in_sync: int,
) -> tuple[
    types.Nested[tf.tpu.experimental.embedding.FeatureConfig],
    tf.tpu.experimental.embedding.SparseCoreEmbeddingConfig,
]:
    """Translates a Keras RS configuration to a TensorFlow TPU configuration.

    Args:
      feature_configs: The nested Keras RS feature configs.
      table_stacking: The Keras RS table stacking.
      num_replicas_in_sync: The number of replicas in sync from the strategy.

    Returns:
      A tuple containing the TensorFlow TPU feature configs and the TensorFlow
      TPU sparse core embedding config.
    """
    tables: dict[int, tf.tpu.experimental.embedding.TableConfig] = {}
    feature_configs = keras.tree.map_structure(
        lambda f: keras_to_tf_tpu_feature_config(
            f, tables, num_replicas_in_sync
        ),
        feature_configs,
    )

    # max_ids_per_chip_per_sample
    # max_ids_per_table
    # max_unique_ids_per_table

    if table_stacking is None:
        disable_table_stacking = True
    elif table_stacking == "auto":
        disable_table_stacking = False
    else:
        raise ValueError(
            f"Unsupported table stacking for Tensorflow {table_stacking}, must "
            "be 'auto' or None."
        )

    # Find alternative.
    # `initialize_tables_on_host` is set to False. Otherwise, if the
    # `TPUEmbedding` layer is built within Keras' `compute_output_spec` (meaning
    # within `call`), the tables are created within a `FuncGraph` and the
    # resulting tables are destroyed at the end of it.
    sparse_core_embedding_config = (
        tf.tpu.experimental.embedding.SparseCoreEmbeddingConfig(
            disable_table_stacking=disable_table_stacking,
            initialize_tables_on_host=False,
        )
    )

    return feature_configs, sparse_core_embedding_config


def keras_to_tf_tpu_feature_config(
    feature_config: FeatureConfig,
    tables: dict[int, tf.tpu.experimental.embedding.TableConfig],
    num_replicas_in_sync: int,
) -> tf.tpu.experimental.embedding.FeatureConfig:
    """Translates a Keras RS feature config to a TensorFlow TPU feature config.

    This creates the table config and adds it to the mapping if it doesn't exist
    in the `tables` mapping`.

    Args:
      feature_config: The Keras RS feature config to translate.
      tables: A mapping of KerasRS table config ids to TF TPU table configs.
      num_replicas_in_sync: The number of replicas in sync from the strategy.

    Returns:
      The TensorFlow TPU feature config.
    """
    if num_replicas_in_sync <= 0:
        raise ValueError(
            "`num_replicas_in_sync` must be positive, "
            f"but got {num_replicas_in_sync}."
        )

    table = tables.get(id(feature_config.table), None)
    if table is None:
        table = keras_to_tf_tpu_table_config(feature_config.table)
        tables[id(feature_config.table)] = table

    if len(feature_config.output_shape) < 2:
        raise ValueError(
            f"Invalid `output_shape` {feature_config.output_shape} in "
            f"`FeatureConfig` {feature_config}. It must have at least 2 "
            "dimensions: a batch dimension and an embedding dimension."
        )

    # Exclude last dimension, TensorFlow's TPUEmbedding doesn't want it.
    output_shape = list(feature_config.output_shape[0:-1])

    batch_size = output_shape[0]
    per_replica_batch_size: int | None = None
    if batch_size is not None:
        if batch_size % num_replicas_in_sync != 0:
            raise ValueError(
                f"Invalid `output_shape` {feature_config.output_shape} in "
                f"`FeatureConfig` {feature_config}. Batch size {batch_size} is "
                f"not a multiple of the number of TPUs {num_replicas_in_sync}."
            )
        per_replica_batch_size = batch_size // num_replicas_in_sync

    # TensorFlow's TPUEmbedding wants the per replica batch size.
    output_shape = [per_replica_batch_size] + output_shape[1:]

    # max_sequence_length
    return tf.tpu.experimental.embedding.FeatureConfig(
        name=feature_config.name,
        table=table,
        output_shape=output_shape,
    )


def keras_to_tf_tpu_table_config(
    table_config: TableConfig,
) -> tf.tpu.experimental.embedding.TableConfig:
    initializer = table_config.initializer
    if isinstance(initializer, str):
        initializer = keras.initializers.get(initializer)

    return tf.tpu.experimental.embedding.TableConfig(
        vocabulary_size=table_config.vocabulary_size,
        dim=table_config.embedding_dim,
        initializer=initializer,
        optimizer=to_tf_tpu_optimizer(table_config.optimizer),
        combiner=table_config.combiner,
        name=table_config.name,
    )


def keras_to_tf_tpu_optimizer(
    optimizer: keras.optimizers.Optimizer,
) -> TfTpuOptimizer:
    """Translates a Keras optimizer to a TensorFlow TPU `_Optimizer`.

    Args:
      optimizer: The Keras optimizer to translate.

    Returns:
      The TensorFlow TPU `_Optimizer`.
    """
    tpu_optimizer_kwargs: dict[str, Any] = {}

    # Supported keras optimizer general options.
    learning_rate = optimizer._learning_rate  # pylint: disable=protected-access
    if isinstance(
        learning_rate, keras.optimizers.schedules.LearningRateSchedule
    ):
        # Note: learning rate requires incrementing iterations in optimizer.
        tpu_optimizer_kwargs["learning_rate"] = lambda: optimizer.learning_rate
    elif callable(learning_rate):
        tpu_optimizer_kwargs["learning_rate"] = learning_rate
    else:
        learning_rate = optimizer.get_config()["learning_rate"]
        if isinstance(learning_rate, float):
            tpu_optimizer_kwargs["learning_rate"] = learning_rate
        else:
            raise ValueError(
                f"Unsupported learning rate: {learning_rate} of type"
                f" {type(learning_rate)}."
            )

    if optimizer.weight_decay is not None:
        tpu_optimizer_kwargs["weight_decay_factor"] = optimizer.weight_decay
    if optimizer.clipvalue is not None:
        tpu_optimizer_kwargs["clipvalue"] = optimizer.clipvalue
    if optimizer.gradient_accumulation_steps is not None:
        tpu_optimizer_kwargs["use_gradient_accumulation"] = True

    # Unsupported keras optimizer general options.
    if optimizer.clipnorm is not None:
        raise ValueError("Unsupported optimizer option `Optimizer.clipnorm`.")
    if optimizer.global_clipnorm is not None:
        raise ValueError(
            "Unsupported optimizer option `Optimizer.global_clipnorm`."
        )
    if optimizer.use_ema:
        raise ValueError("Unsupported optimizer option `Optimizer.use_ema`.")
    if optimizer.loss_scale_factor is not None:
        raise ValueError(
            "Unsupported optimizer option `Optimizer.loss_scale_factor`."
        )

    optimizer_mapping = None
    for optimizer_class, mapping in OPTIMIZER_MAPPINGS.items():
        # Handle subclasses of the main optimizer class.
        if isinstance(optimizer, optimizer_class):
            optimizer_mapping = mapping
            break
    if optimizer_mapping is None:
        raise ValueError(
            f"Unsupported optimizer type {type(optimizer)}. Optimizer must be "
            f"one of {list(OPTIMIZER_MAPPINGS.keys())}."
        )

    for argname in optimizer_mapping.supported_kwargs:
        tpu_optimizer_kwargs[argname] = getattr(optimizer, argname)

    for argname, disabled_value in optimizer_mapping.unsupported_kwargs.items():
        if disabled_value is None:
            if getattr(optimizer, argname) is not None:
                raise ValueError(f"Unsupported optimizer option {argname}.")
        elif getattr(optimizer, argname) != disabled_value:
            raise ValueError(f"Unsupported optimizer option {argname}.")

    return optimizer_mapping.tpu_optimizer_class(**tpu_optimizer_kwargs)


def to_tf_tpu_optimizer(
    optimizer: str | keras.optimizers.Optimizer | TfTpuOptimizer | None,
) -> TfTpuOptimizer:
    """Translates a Keras optimizer into a TensorFlow TPU `_Optimizer`.

    Args:
      optimizer: The optimizer to translate.

    Returns:
      The equivalent TensorFlow TPU `_Optimizer`.

    Raises:
      ValueError: If the optimizer or one of its argument is not supported.
    """
    if optimizer is None:
        return None
    elif isinstance(
        optimizer,
        (
            tf.tpu.experimental.embedding.SGD,
            tf.tpu.experimental.embedding.Adagrad,
            tf.tpu.experimental.embedding.Adam,
            tf.tpu.experimental.embedding.FTRL,
        ),
    ):
        return optimizer
    elif isinstance(optimizer, str):
        if optimizer == "sgd":
            return tf.tpu.experimental.embedding.SGD()
        elif optimizer == "adagrad":
            return tf.tpu.experimental.embedding.Adagrad()
        elif optimizer == "adam":
            return tf.tpu.experimental.embedding.Adam()
        elif optimizer == "ftrl":
            return tf.tpu.experimental.embedding.FTRL()
        else:
            raise ValueError(
                f"Unknown optimizer name '{optimizer}'. Please use one of "
                "'sgd', 'adagrad', 'adam', or 'ftrl'"
            )
    elif isinstance(optimizer, keras.optimizers.Optimizer):
        return keras_to_tf_tpu_optimizer(optimizer)
    else:
        raise ValueError(
            f"Unknown optimizer type {type(optimizer)}. Please pass an "
            "optimizername as a string, a subclass of keras optimizer or an "
            "instance of one of the optimizer parameter classes in "
            "`tf.tpu.experimental.embedding`."
        )


# TensorFlow to TensorFlow


def clone_tf_tpu_feature_configs(
    feature_configs: types.Nested[tf.tpu.experimental.embedding.FeatureConfig],
) -> types.Nested[tf.tpu.experimental.embedding.FeatureConfig]:
    """Clones and resolves TensorFlow TPU feature configs.

    This function clones the feature configs and resolves the table configs.

    Args:
      feature_configs: The TensorFlow TPU feature configs to clone and resolve.

    Returns:
      The cloned and resolved TensorFlow TPU feature configs.
    """
    table_configs_dict = {}

    def clone_and_resolve_tf_tpu_feature_config(
        fc: tf.tpu.experimental.embedding.FeatureConfig,
    ) -> tf.tpu.experimental.embedding.FeatureConfig:
        if fc.table not in table_configs_dict:
            table_configs_dict[fc.table] = (
                tf.tpu.experimental.embedding.TableConfig(
                    vocabulary_size=fc.table.vocabulary_size,
                    dim=fc.table.dim,
                    initializer=fc.table.initializer,
                    optimizer=to_tf_tpu_optimizer(fc.table.optimizer),
                    combiner=fc.table.combiner,
                    name=fc.table.name,
                    quantization_config=fc.table.quantization_config,
                    layout=fc.table.layout,
                )
            )
        return tf.tpu.experimental.embedding.FeatureConfig(
            table=table_configs_dict[fc.table],
            max_sequence_length=fc.max_sequence_length,
            validate_weights_and_indices=fc.validate_weights_and_indices,
            output_shape=fc.output_shape,
            name=fc.name,
        )

    return keras.tree.map_structure(
        clone_and_resolve_tf_tpu_feature_config, feature_configs
    )



"""Configuration for TPU embedding layer."""

import dataclasses
from typing import Any

import keras

from keras_rs.src import types
from keras_rs.src.api_export import keras_rs_export


@keras_rs_export("keras_rs.layers.TableConfig")
@dataclasses.dataclass(order=True)
class TableConfig:
    """Configuration for one embedding table.

    Configures one table for use by one or more `keras_rs.layers.FeatureConfig`,
    which in turn is used to configure a `keras_rs.layers.DistributedEmbedding`.

    Attributes:
        name: The name of the table. Must be defined.
        vocabulary_size: Size of the table's vocabulary (number of rows).
        embedding_dim: The embedding dimension (width) of the table.
        initializer: The initializer for the embedding weights. If not
            specified,  defaults to `truncated_normal_initializer` with mean
            `0.0` and standard deviation `1 / sqrt(embedding_dim)`.
        optimizer: The optimizer for the embedding table. Only SGD, Adagrad,
            Adam, and FTRL are supported. Note that not all of the optimizer's
            parameters are supported. Defaults to Adam.
        combiner: Specifies how to reduce if there are multiple entries in a
            single row. `mean`, `sqrtn` and `sum` are supported. `mean` is the
            default. `sqrtn` often achieves good accuracy, in particular with
            bag-of-words columns.
        placement: Where to place the embedding table. `"auto"`, which is the
            default, means that the table is placed on SparseCore if available,
            otherwise on the default device where the rest of the model is
            placed. A value of `"sparsecore"` means the table will be placed on
            the SparseCore chips and an error is raised if SparseCore is not
            available. A value of `"default_device"` means the table will be
            placed on the default device where the rest of the model is placed,
            even if SparseCore is available. The default device for the rest of
            the model is the TPU's TensorCore on TPUs, otherwise the GPU or CPU.
        max_ids_per_partition: The max number of ids per partition for the
            table. This is an input data dependent value and is required by the
            compiler to appropriately allocate memory.
        max_unique_ids_per_partition: The max number of unique ids per partition
            for the table. This is an input data dependent value and is required
            by the compiler to appropriately allocate memory.
    """

    name: str
    vocabulary_size: int
    embedding_dim: int
    initializer: str | keras.initializers.Initializer = (
        keras.initializers.VarianceScaling(mode="fan_out")
    )
    optimizer: str | keras.optimizers.Optimizer = "adam"
    combiner: str = "mean"
    placement: str = "auto"
    max_ids_per_partition: int = 256
    max_unique_ids_per_partition: int = 256

    def get_config(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "vocabulary_size": self.vocabulary_size,
            "embedding_dim": self.embedding_dim,
            "initializer": keras.saving.serialize_keras_object(
                self.initializer
            ),
            "optimizer": keras.saving.serialize_keras_object(self.optimizer),
            "combiner": self.combiner,
            "placement": self.placement,
            "max_ids_per_partition": self.max_ids_per_partition,
            "max_unique_ids_per_partition": self.max_unique_ids_per_partition,
        }

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "TableConfig":
        config = config.copy()
        config["initializer"] = keras.saving.deserialize_keras_object(
            config["initializer"]
        )
        config["optimizer"] = keras.saving.deserialize_keras_object(
            config["optimizer"]
        )
        return cls(**config)


@keras_rs_export("keras_rs.layers.FeatureConfig")
@dataclasses.dataclass(order=True)
class FeatureConfig:
    """Configuration for one embedding feature.

    Configures one feature for `keras_rs.layers.DistributedEmbedding`. Each
    feature uses a table configured via `keras_rs.layers.TableConfig` and
    multiple features can share the same table.

    Attributes:
        name: The name of the feature. Must be defined.
        table: The table in which to look up this feature.
        input_shape: The input shape of the feature. The feature fed into the
            layer has to match the shape. Note that for ragged dimensions in the
            input, the dimension provided here presents the maximum value;
            anything larger will be truncated. Also note that the first
            dimension represents the global batch size. For example, on TPU,
            this represents the total number of samples that are dispatched to
            all the TPUs connected to the current host.
        output_shape: The output shape of the feature activation. What is
            returned by the embedding layer has to match this shape.
    """

    name: str
    table: TableConfig
    input_shape: types.Shape
    output_shape: types.Shape

    def get_config(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "table": self.table.get_config(),
            "input_shape": self.input_shape,
            "output_shape": self.output_shape,
        }

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "FeatureConfig":
        config = config.copy()
        # Note: the handling of shared tables during serialization is done in
        # `DistributedEmbedding.from_config()`.
        config["table"] = TableConfig.from_config(config["table"])
        return cls(**config)



"""A Wrapper over orbax CheckpointManager for Keras3 Jax TPU Embeddings."""

from typing import Any

import keras
import orbax.checkpoint as ocp
from etils import epath


class JaxKeras3CheckpointManager(ocp.CheckpointManager):
    """A wrapper over orbax CheckpointManager for Keras3 Jax TPU Embeddings."""

    def __init__(
        self,
        model: keras.Model,
        checkpoint_dir: epath.PathLike,
        max_to_keep: int,
        steps_per_epoch: int = 1,
        **kwargs: Any,
    ):
        options = ocp.CheckpointManagerOptions(
            max_to_keep=max_to_keep, enable_async_checkpointing=False, **kwargs
        )
        self._model = model
        self._steps_per_epoch = steps_per_epoch
        self._checkpoint_dir = checkpoint_dir
        super().__init__(checkpoint_dir, options=options)

    def _get_state(self) -> tuple[dict[str, Any], Any | None]:
        """Gets the model state and metrics"""
        model_state = self._model.get_state_tree()
        state = {}
        metrics = None
        for k, v in model_state.items():
            if k == "metrics_variables":
                metrics = v
            else:
                state[k] = v
        return state, metrics

    def save_state(self, epoch: int) -> None:
        """Saves the model to the checkpoint directory.

        Args:
          epoch: The epoch number at which the state is saved.
        """
        state, metrics_value = self._get_state()
        self.save(
            epoch * self._steps_per_epoch,
            args=ocp.args.StandardSave(item=state),
            metrics=metrics_value,
        )

    def restore_state(self, step: int | None = None) -> None:
        """Restores the model from the checkpoint directory.

        Args:
          step: The step .number to restore the state from. Default=None
            restores the latest step.
        """
        if step is None:
            step = self.latest_step()
        # Restore the model state only, not metrics.
        state, _ = self._get_state()
        restored_state = self.restore(
            step, args=ocp.args.StandardRestore(item=state)
        )
        self._model.set_state_tree(restored_state)


class JaxKeras3CheckpointCallback(keras.callbacks.Callback):
    """A callback for checkpointing and restoring state using Orbax."""

    def __init__(
        self,
        model: keras.Model,
        checkpoint_dir: epath.PathLike,
        max_to_keep: int,
        steps_per_epoch: int = 1,
        **kwargs: Any,
    ):
        if keras.backend.backend() != "jax":
            raise ValueError(
                "`JaxKeras3CheckpointCallback` is only supported on a "
                "`jax` backend."
            )
        self._checkpoint_manager = JaxKeras3CheckpointManager(
            model, checkpoint_dir, max_to_keep, steps_per_epoch, **kwargs
        )

    def on_train_begin(self, logs: dict[str, Any] | None = None) -> None:
        if not self.model.built or not self.model.optimizer.built:
            raise ValueError(
                "To use `JaxKeras3CheckpointCallback`, your model and "
                "optimizer must be built before you call `fit()`."
            )
        latest_epoch = self._checkpoint_manager.latest_step()
        if latest_epoch is not None:
            self._checkpoint_manager.restore_state(step=latest_epoch)

    def on_epoch_end(
        self, epoch: int, logs: dict[str, Any] | None = None
    ) -> None:
        self._checkpoint_manager.save_state(epoch)



"""Type definitions."""

from typing import Any, Callable, Mapping, Sequence, TypeAlias, TypeVar, Union

import keras

"""
A tensor in any of the backends.

We do not define it explicitly to not require all the backends to be installed
and imported. The explicit definition would be:
```
numpy.ndarray,
| tensorflow.Tensor,
| tensorflow.RaggedTensor,
| tensorflow.SparseTensor,
| tensorflow.IndexedSlices,
| jax.Array,
| jax.experimental.sparse.JAXSparse,
| torch.Tensor,
| keras.KerasTensor,
```
"""
Tensor: TypeAlias = Any

Shape: TypeAlias = Sequence[int | None]

DType: TypeAlias = str

ConstraintLike: TypeAlias = (
    str
    | keras.constraints.Constraint
    | type[keras.constraints.Constraint]
    | Callable[[Tensor], Tensor]
)

InitializerLike: TypeAlias = (
    str
    | keras.initializers.Initializer
    | type[keras.initializers.Initializer]
    | Callable[[Shape, DType], Tensor]
    | Tensor
)

RegularizerLike: TypeAlias = (
    str
    | keras.regularizers.Regularizer
    | type[keras.regularizers.Regularizer]
    | Callable[[Tensor], Tensor]
)

T = TypeVar("T")
Nested: TypeAlias = (
    T | Sequence[Union[T, "Nested[T]"]] | Mapping[str, Union[T, "Nested[T]"]]
)
